from __future__ import annotations

import numpy as np
import numpy.typing as npt
from astropy import units
from astropy.modeling.physical_models import BlackBody
from scipy import integrate

MIN_TEMP = 40 * units.K
MAX_TEMP = 550 * units.K
N_TEMPS = 100
TEMPERATURES = np.linspace(MIN_TEMP, MAX_TEMP, N_TEMPS)
blackbody = BlackBody(TEMPERATURES)


def get_dust_grain_temperature(
    R: npt.NDArray[np.float64], T_0: float, delta: float
) -> npt.NDArray[np.float64]:
    """Return the dust grain temperature given at a radial distance from the Sun.

    Args:
        R: Radial distance from the sun in ecliptic heliocentric coordinates [AU / 1AU].
        T_0: Temperature of dust grains located 1 AU from the Sun [K].
        delta: Powerlaw index.

    Returns:
        Dust grain temperature [K].

    """
    return T_0 * R**-delta


def tabulate_blackbody_emission(
    wavelengths: units.Quantity, weights: units.Quantity | None
) -> npt.NDArray[np.float64]:
    """Tabulate bandpass integrated blackbody specific intensity for a range of temperatures."""
    if weights is None:
        tabulated_blackbody_emission = blackbody(wavelengths)
    else:
        tabulated_blackbody_emission = integrate.trapezoid(
            weights * blackbody(wavelengths[:, np.newaxis]).transpose(), wavelengths
        )

    return np.asarray(
        [
            TEMPERATURES.to_value(units.K),
            tabulated_blackbody_emission.to_value(units.MJy / units.sr),
        ]
    )



from __future__ import annotations

import functools
import itertools
import multiprocessing
import platform

import numpy as np
import numpy.typing as npt
from astropy import coordinates as coords
from astropy import time, units
from scipy import integrate

from zodipy.blackbody import tabulate_blackbody_emission
from zodipy.bodies import (
    get_earthpos_inst,
    get_interp_obstimes,
    get_interpolated_bodypos,
    get_obspos_from_str,
)
from zodipy.component import ComponentLabel
from zodipy.line_of_sight import (
    get_line_of_sight_range_dicts,
    integrate_leggauss,
)
from zodipy.model_registry import model_registry
from zodipy.number_density import get_partial_number_density_func, update_partial_earth_pos
from zodipy.unpack_model import get_model_interp_func

PLATFORM_METHOD = "fork" if "windows" not in platform.system().lower() else None


class Model:
    """Main interface to ZodiPy."""

    def __init__(
        self,
        x: units.Quantity,
        *,
        weights: npt.ArrayLike | None = None,
        name: str = "dirbe",
        gauss_quad_degree: int = 50,
        extrapolate: bool = False,
        ephemeris: str = "builtin",
    ) -> None:
        """Initialize the ZodiPy model interface.

        Args:
            x: Wavelength or frequency. If `x` is a sequence, it is assumed to be a the points
                corresponding to a bandpass and the corresponding `weights` must be provided.
            weights: Bandpass weights corresponding the the frequencies or wavelengths in `x`. The
                weights are assumed to represent a normalized instrument response in units of
                spectral radiance (Jy/sr).
            name: Zodiacal light model to use for the simulations. For a list of available models,
                see https://cosmoglobe.github.io/zodipy/introduction/. Defaults to 'dirbe'.
            gauss_quad_degree: Order of the Gaussian-legendre quadrature used to evaluate the
                line-of-sight integrals in the simulations. Default is 50 points.
            extrapolate: If `True` all spectral quantities in the selected model are extrapolated to
                the requested frequencies or wavelengths. If `False`, an exception is raised on
                requested values of `x` outside of the valid model range. Default is `False`.
            ephemeris: Ephemeris used in `astropy.coordinates.solar_system_ephemeris` to compute the
                positions of the observer and Earth. Defaults to 'builtin'. See the
                [Astropy documentation](https://docs.astropy.org/en/stable/coordinates/solarsystem.html)
                for available ephemerides.

        """
        try:
            if not x.isscalar and weights is None:
                msg = "Bandpass weights must be provided for non-scalar `x`."
                raise ValueError(msg)
        except AttributeError as error:
            msg = "The input 'x' must be an astropy Quantity."
            raise TypeError(msg) from error
        if x.isscalar and weights is not None:
            msg = "Bandpass weights should not be provided for scalar `x`."
            raise ValueError(msg)

        self._ipd_model = model_registry.get_model(name)

        if not extrapolate and not self._ipd_model.is_valid_at(x):
            msg = (
                "The requested frequencies are outside the valid range of the model. "
                "If this was intended, set the extrapolate argument to True."
            )
            raise ValueError(msg)

        # Bandpass is provided rather than a delta wavelength or frequency.
        if weights is not None:
            weights = np.asarray(weights)
            if x.size != weights.size:
                msg = "Number of wavelengths and weights must be the same in the bandpass."
                raise ValueError(msg)
            normalized_weights = weights / integrate.trapezoid(weights, x)
        else:
            normalized_weights = None

        self._x = x
        self._normalized_weights = normalized_weights
        self._b_nu_table = tabulate_blackbody_emission(self._x, self._normalized_weights)

        quad_points, quad_weights = np.polynomial.legendre.leggauss(gauss_quad_degree)
        self._integrate_leggauss = functools.partial(
            integrate_leggauss,
            points=quad_points,
            weights=quad_weights,
        )

        self._ephemeris = ephemeris

        # Make mypy happy by declaring types of to-be initialized attributes.
        self._number_density_partials: dict[ComponentLabel, functools.partial]
        self._interped_comp_params: dict[ComponentLabel, dict]
        self._interped_shared_params: dict

        self._init_ipd_model_partials()

    def evaluate(
        self,
        skycoord: coords.SkyCoord,
        *,
        obspos: units.Quantity | str = "earth",
        return_comps: bool = False,
        nprocesses: int = 1,
    ) -> units.Quantity[units.MJy / units.sr]:
        """Return the simulated zodiacal light.

        The zodiacal light is simulated for a single, or a sequence of observations. If a single
        `obspos` and `obstime` is provided for multiple coordinates, all coordinates are assumed to
        be observed from that position at that time. Otherwise, each coordinate is simulated from
        the corresponding observer position and time.

        Args:
            skycoord: `astropy.coordinates.SkyCoord` object representing the coordinates or
                observations for which to simulate the zodiacal light. The `frame` and `obstime`
                attributes of the `SkyCoord` object must be set. The `obstime` attribute must be
                specified, and correspond to a single, or a sequence of observational times with
                length matching the number of coordinates. The frame must be convertible to the
                `astropy.coordinates.BarycentricMeanEcliptic` frame.
            obspos: The heliocentric ecliptic position of the observer, or a string representing
                an observer in the `astropy.coordinates.solar_system_ephemeris`. If an explicit
                position is given, it must either be a single, or a sequence of positions with
                shape matching the number of coordinates Defaults to 'earth'.
            return_comps: If True, the emission is returned component-wise. Defaults to False.
            nprocesses: Number of cores to use. If `nprocesses >= 1`, the line-of-sight integrals
                are parallelized using the `multiprocessing` module. Defaults to 1.

        Returns:
            emission: Simulated zodiacal light in units of 'MJy/sr'.

        """
        try:
            if skycoord.obstime is None:
                msg = "The `obstime` attribute of the `SkyCoord` object is not set."
                raise ValueError(msg)
        except AttributeError as error:
            msg = "The input coordinates must be an astropy SkyCoord object."
            raise TypeError(msg) from error

        try:
            if not (obspos_isstr := isinstance(obspos, str)) and (
                (obspos.ndim > 1 and skycoord.obstime.size != obspos.shape[-1])
                or (obspos.ndim == 1 and skycoord.obstime.size != 1)
            ):
                msg = "The number of obstime (ncoords) and obspos (3, ncoords) does not match."
                raise ValueError(msg)
        except AttributeError as error:
            msg = "The observer position is not a string or an astropy Quantity."
            raise TypeError(msg) from error

        if skycoord.obstime.size > skycoord.size:
            msg = "The size of obstime must be either 1 or ncoords."
            raise ValueError(msg)

        if skycoord.obstime.size == 1:
            interp_obstimes = None
        else:
            interp_obstimes = get_interp_obstimes(skycoord.obstime[0].mjd, skycoord.obstime[-1].mjd)

        dist_coords_to_cores = skycoord.size > nprocesses and nprocesses > 1
        if dist_coords_to_cores:
            skycoord_splits = np.array_split(skycoord, nprocesses)
            obspos_splits = (
                itertools.repeat(obspos, nprocesses)
                if obspos_isstr
                else np.array_split(obspos, nprocesses, axis=-1)
            )
            interp_obstime_splits = itertools.repeat(interp_obstimes, nprocesses)
            with multiprocessing.get_context(PLATFORM_METHOD).Pool(nprocesses) as pool:
                emission_splits = [
                    pool.apply_async(self._evaluate, args=(skycoord, obspos, obstime_lims))
                    for skycoord, obspos, obstime_lims in zip(
                        skycoord_splits, obspos_splits, interp_obstime_splits
                    )
                ]
                emission = np.concatenate([split.get() for split in emission_splits], axis=-1)
        else:
            emission = self._evaluate(skycoord, obspos, interp_obstimes)

        emission <<= units.MJy / units.sr
        return emission if return_comps else emission.sum(axis=0)

    def _evaluate(
        self,
        skycoord: coords.SkyCoord,
        obspos: units.Quantity | str,
        interp_obstimes: time.Time | None,
    ) -> npt.NDArray[np.float64]:
        """Evaluate the zodiacal light for a single or a sequence of sky coordinates."""
        if interp_obstimes is None:
            earth_xyz = get_earthpos_inst(skycoord.obstime, self._ephemeris)
        else:
            earth_xyz = get_interpolated_bodypos(
                body="earth",
                obstimes=skycoord.obstime.mjd,
                interp_obstimes=interp_obstimes,
                ephemeris=self._ephemeris,
            )

        if isinstance(obspos, str):
            obs_xyz = get_obspos_from_str(
                body=obspos,
                obstime=skycoord.obstime,
                interp_obstimes=interp_obstimes,
                earthpos=earth_xyz,
                ephemeris=self._ephemeris,
            )
        else:
            try:
                obs_xyz = obspos.to_value(units.AU)
            except units.UnitConversionError as error:
                msg = "The observer position must be in length units."
                raise units.UnitConversionError(msg) from error

        instantaneous = skycoord.obstime.size == 1
        if instantaneous:
            obs_xyz = obs_xyz[:, np.newaxis]
        if earth_xyz.ndim == 1:
            earth_xyz = earth_xyz[:, np.newaxis]

        # Model evaluation is performed in heliocentric ecliptic coordinates. We transform
        # to the barycentric frame, which is compatiable with the Galactic and Celestial,
        # and pretend that that this is the heliocentric frame as we only need the correction
        # rotation.
        skycoord = skycoord.transform_to(coords.BarycentricMeanEcliptic)

        skycoord_xyz: npt.NDArray[np.float64] = skycoord.cartesian.xyz.value
        if skycoord.isscalar:
            skycoord_xyz = skycoord.cartesian.xyz.value[:, np.newaxis]

        start, stop = get_line_of_sight_range_dicts(
            components=self._ipd_model.comps.keys(),
            unit_vectors=skycoord_xyz,
            obs_pos=obs_xyz,
        )

        number_density_partials = self._number_density_partials
        shared_brightness_partial = self._shared_brightness_partial

        number_density_partials = update_partial_earth_pos(
            number_density_partials, earth_pos=earth_xyz
        )
        shared_brightness_partial = functools.partial(shared_brightness_partial, X_obs=obs_xyz)

        emission = np.zeros((self._ipd_model.ncomps, skycoord.size))
        for idx, comp_label in enumerate(self._ipd_model.comps.keys()):
            comp_func = functools.partial(
                shared_brightness_partial,
                u_los=skycoord_xyz,
                start=start[comp_label],
                stop=stop[comp_label],
                number_density_func=number_density_partials[comp_label],
                **self._interped_comp_params[comp_label],
            )
            emission[idx] = self._integrate_leggauss(comp_func)

        return emission

    def _init_ipd_model_partials(self) -> None:
        """Initialize the partial functions for the interplanetary dust model.

        The spectrally dependant model parameters are interpolated over the provided bandpass or
        delta frequency/wavelength. The partial functions are pre-populated functions that contains
        all non line-of-sight related parameters.
        """
        interp_and_unpack_func = get_model_interp_func(self._ipd_model)
        dicts = interp_and_unpack_func(self._x, self._normalized_weights, self._ipd_model)
        self._interped_comp_params = dicts[0]
        self._interped_shared_params = dicts[1]

        self._shared_brightness_partial = functools.partial(
            self._ipd_model.brightness_at_step_callable,
            bp_interpolation_table=self._b_nu_table,
            **self._interped_shared_params,
        )

        self._number_density_partials = get_partial_number_density_func(comps=self._ipd_model.comps)

    def get_parameters(self) -> dict:
        """Return a dictionary containing the interplanetary dust model parameters.

        This method is mainly meant to be used to fit or sample zodiacal light models.

        Returns:
            parameters: Dictionary of parameters of the interplanetary dust model.
        """
        return self._ipd_model.to_dict()

    def update_parameters(self, parameters: dict) -> None:
        """Update the interplanetary dust model parameters.

        This method is mainly meant to be used to fit or sample zodiacal light models.

        Args:
            parameters: Dictionary of parameters to update. The keys must be the names
                of the parameters as defined in the model. To get the parameters dict
                of an existing model, use the`get_parameters` method of an initialized
                `zodipy.Model`.
        """
        _dict = parameters.copy()
        _dict["comps"] = {}
        for key, value in parameters.items():
            if key == "comps":
                for comp_key, comp_value in value.items():
                    _dict["comps"][ComponentLabel(comp_key)] = type(
                        self._ipd_model.comps[ComponentLabel(comp_key)]
                    )(**comp_value)
            elif isinstance(value, dict):
                _dict[key] = {ComponentLabel(k): v for k, v in value.items()}

        self._ipd_model = self._ipd_model.__class__(**_dict)
        self._init_ipd_model_partials()



from __future__ import annotations

from typing import TYPE_CHECKING

import astropy.coordinates as coords
import numpy as np
from astropy import time, units
from scipy import interpolate

if TYPE_CHECKING:
    import numpy.typing as npt

MEAN_DIST_TO_L2 = 0.009896235034000056


def get_interp_obstimes(t0: float, t1: float) -> time.Time:
    """Return a subset of the obstimes used to interpolate in body positions."""
    dt = (1 * units.hour).to_value(units.day)
    return time.Time(np.arange(t0, t1 + dt, dt), format="mjd")


def get_interpolated_bodypos(
    body: str,
    obstimes: npt.NDArray[np.float64],
    interp_obstimes: time.Time,
    ephemeris: str,
) -> np.ndarray:
    """Return the interpolated heliocentric positions of body."""
    pos = (
        coords.get_body(body, interp_obstimes, ephemeris=ephemeris)
        .transform_to(coords.HeliocentricMeanEcliptic)
        .cartesian.xyz.to_value(units.AU)
    )
    interpolator = interpolate.CubicSpline(interp_obstimes.mjd, pos, axis=-1)
    return interpolator(obstimes)


def get_semb_l2_pos(
    earthpos: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """Return a SkyCoord of the heliocentric position of the SEMB-L2 point.

    Note that this is an approximate position, as the SEMB-L2 point is not included in
    any of the current available ephemerides. We assume that SEMB-L2 is at all times
    located at a fixed distance from Earth along the vector pointing to Earth from the Sun.
    """
    earth_distance = np.linalg.norm(earthpos)
    SEMB_L2_distance = earth_distance + MEAN_DIST_TO_L2
    earth_unit_vector = earthpos / earth_distance
    return earth_unit_vector * SEMB_L2_distance


def get_earthpos_inst(
    obstime: time.Time,
    ephemeris: str,
) -> npt.NDArray[np.float64]:
    """Return the sky coordinates of the Earth in the heliocentric frame."""
    return (
        coords.get_body("earth", obstime, ephemeris=ephemeris)
        .transform_to(coords.HeliocentricMeanEcliptic)
        .cartesian.xyz.to_value(units.AU)
    ).flatten()


def get_obspos_from_str(
    body: str,
    obstime: time.Time,
    interp_obstimes: time.Time | None,
    earthpos: npt.NDArray[np.float64],
    ephemeris: str,
) -> npt.NDArray[np.float64]:
    """Return the sky coordinates of the observer in the heliocentric frame."""
    if body == "semb-l2":
        return get_semb_l2_pos(earthpos)
    if body == "earth":
        return earthpos

    if obstime.size == 1:
        try:
            return (
                coords.get_body(body, obstime, ephemeris=ephemeris)
                .transform_to(coords.HeliocentricMeanEcliptic)
                .cartesian.xyz.to_value(units.AU)
            ).flatten()
        except KeyError as error:
            valid_obs = [*coords.solar_system_ephemeris.bodies, "semb-l2"]
            msg = f"Invalid observer string: '{body}'. Valid observers are: {valid_obs}"
            raise ValueError(msg) from error

    if interp_obstimes is None:  # pragma: no cover
        msg = "interp_obstimes must be provided when obstime is an array."
        raise ValueError(msg)

    return get_interpolated_bodypos(
        body=body,
        obstimes=obstime.mjd,
        interp_obstimes=interp_obstimes,
        ephemeris=ephemeris,
    )



from zodipy import component_params, source_params
from zodipy.zodiacal_light_model import RRM, Kelsall, model_registry

model_registry.register_model(
    name="dirbe",
    model=Kelsall(
        comps=component_params.DIRBE,
        spectrum=source_params.SPECTRUM_DIRBE,
        emissivities=source_params.EMISSIVITY_DIRBE,
        albedos=source_params.ALBEDO_DIRBE,
        solar_irradiance=source_params.SOLAR_IRRADIANCE_DIRBE,
        C1=source_params.C1_DIRBE,
        C2=source_params.C2_DIRBE,
        C3=source_params.C3_DIRBE,
        T_0=source_params.T_0_DIRBE,
        delta=source_params.DELTA_DIRBE,
    ),
)

model_registry.register_model(
    name="planck13",
    model=Kelsall(
        comps=component_params.DIRBE,
        spectrum=source_params.SPECTRUM_PLANCK,
        emissivities=source_params.EMISSIVITY_PLANCK_13,
        T_0=source_params.T_0_DIRBE,
        delta=source_params.DELTA_DIRBE,
    ),
)

model_registry.register_model(
    name="planck15",
    model=Kelsall(
        comps=component_params.PLANCK,
        spectrum=source_params.SPECTRUM_PLANCK,
        emissivities=source_params.EMISSIVITY_PLANCK_15,
        T_0=source_params.T_0_DIRBE,
        delta=source_params.DELTA_DIRBE,
    ),
)

model_registry.register_model(
    name="planck18",
    model=Kelsall(
        comps=component_params.PLANCK,
        spectrum=source_params.SPECTRUM_PLANCK,
        emissivities=source_params.EMISSIVITY_PLANCK_18,
        T_0=source_params.T_0_DIRBE,
        delta=source_params.DELTA_DIRBE,
    ),
)

model_registry.register_model(
    name="odegard",
    model=Kelsall(
        comps=component_params.PLANCK,
        spectrum=source_params.SPECTRUM_PLANCK,
        emissivities=source_params.EMISSIVITY_ODEGARD,
        T_0=source_params.T_0_DIRBE,
        delta=source_params.DELTA_DIRBE,
    ),
)

model_registry.register_model(
    name="rrm-experimental",
    model=RRM(
        comps=component_params.RRM,
        spectrum=source_params.SPECTRUM_IRAS,
        calibration=source_params.CALIBRATION_RRM,
        T_0=source_params.T_0_RRM,
        delta=source_params.DELTA_RMM,
    ),
)



from __future__ import annotations

import abc
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Mapping, Sequence

import numpy as np
from astropy import units

from zodipy.brightness import (
    BrightnessAtStepCallable,
    kelsall_brightness_at_step,
    rrm_brightness_at_step,
)

if TYPE_CHECKING:
    from zodipy.component import ComponentLabel, ZodiacalComponent


@dataclass(repr=False)
class ZodiacalLightModel(abc.ABC):
    """Base class for interplanetary dust models.

    Args:
        comps: Mapping of component labels to component classes.
        spectrum: Frequency or wavelength over which the model is valid.

    """

    comps: Mapping[ComponentLabel, ZodiacalComponent]
    spectrum: units.Quantity

    @property
    @abc.abstractmethod
    def brightness_at_step_callable(cls) -> BrightnessAtStepCallable:
        """Return the callable that computes the brightness at a step."""

    def to_dict(self) -> dict:
        """Return a dictionary representation of the model."""
        _dict: dict = {}
        for key, value in vars(self).items():
            if key == "comps":
                _dict[key] = {}
                for comp_key, comp_value in value.items():
                    _dict[key][comp_key.value] = {
                        k: v
                        for k, v in vars(comp_value).items()
                        if comp_value.__dataclass_fields__[k].init
                    }
            elif isinstance(value, dict):
                _dict[key] = {k.value: v for k, v in value.items()}
            else:
                _dict[key] = value

        return _dict

    @property
    def ncomps(self) -> int:
        """Return the number of components in the model."""
        return len(self.comps)

    def is_valid_at(self, wavelengths: units.Quantity) -> np.bool_:
        """Check if the model is valid at a given wavelength."""
        try:
            wavelengths = wavelengths.to(self.spectrum.unit, equivalencies=units.spectral())
        except units.UnitConversionError as error:
            msg = "Input 'x' must have units convertible to Hz or m."
            raise units.UnitConversionError(msg) from error
        return np.all((self.spectrum.min() <= wavelengths) & (wavelengths <= self.spectrum.max()))


@dataclass(repr=False)
class Kelsall(ZodiacalLightModel):
    """Kelsall et al. (1998) model."""

    T_0: float
    delta: float
    emissivities: Mapping[ComponentLabel, Sequence[float]]
    albedos: Mapping[ComponentLabel, Sequence[float]] | None = None
    solar_irradiance: Sequence[float] | None = None  # In units of MJy/sr
    C1: Sequence[float] | None = None
    C2: Sequence[float] | None = None
    C3: Sequence[float] | None = None

    @property
    def brightness_at_step_callable(cls) -> BrightnessAtStepCallable:
        """Kellsall brightness at a step fuction."""
        return kelsall_brightness_at_step


@dataclass(repr=False)
class RRM(ZodiacalLightModel):
    """Rowan-Robinson and May (2013) model."""

    T_0: Mapping[ComponentLabel, float]
    delta: Mapping[ComponentLabel, float]
    calibration: Sequence[float]

    @property
    def brightness_at_step_callable(cls) -> BrightnessAtStepCallable:
        """RRM brightness at a step fuction."""
        return rrm_brightness_at_step


@dataclass
class ModelRegistry:
    """Container for registered models."""

    _registry: dict[str, ZodiacalLightModel] = field(init=False, default_factory=dict)

    @property
    def models(self) -> list[str]:
        """Return a list of registered models."""
        return list(self._registry.keys())

    def register_model(
        self,
        name: str,
        model: ZodiacalLightModel,
    ) -> None:
        """Register a model with the registry."""
        if (name := name.lower()) in self._registry:
            msg = f"a model by the name {name!s} is already registered."
            raise ValueError(msg)
        if not isinstance(model, ZodiacalLightModel):  # pragma: no cover
            msg = "model must be an instance of ZodiacalLightModel."
            raise TypeError(msg)
        self._registry[name] = model

    def get_model(self, name: str) -> ZodiacalLightModel:
        """Return a model from the registry."""
        if (name := name.lower()) not in self._registry:
            msg = (
                f"{name!r} is not a registered Interplanetary Dust model. "
                f"Avaliable models are: {', '.join(self._registry)}."
            )
            raise ValueError(msg)

        return self._registry[name]


model_registry = ModelRegistry()
